<?php

namespace SGalinski\SgMail\Controller;

/***************************************************************
 *  Copyright notice
 *
 *  (c) sgalinski Internet Services (https://www.sgalinski.de)
 *
 *  All rights reserved
 *
 *  This script is part of the TYPO3 project. The TYPO3 project is
 *  free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  The GNU General Public License can be found at
 *  http://www.gnu.org/copyleft/gpl.html.
 *
 *  This script is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  This copyright notice MUST APPEAR in all copies of the script!
 ***************************************************************/

use SGalinski\SgMail\Domain\Model\Template;
use SGalinski\SgMail\Domain\Repository\FrontendUserGroupRepository;
use SGalinski\SgMail\Domain\Repository\LayoutRepository;
use SGalinski\SgMail\Domain\Repository\TemplateRepository;
use SGalinski\SgMail\Service\BackendService;
use SGalinski\SgMail\Service\MailTemplateService;
use SGalinski\SgMail\Service\RegisterService;
use SGalinski\SgMail\Session\PhpSession;
use TYPO3\CMS\Backend\Template\Components\DocHeaderComponent;
use TYPO3\CMS\Backend\Utility\BackendUtility;
use TYPO3\CMS\Core\Messaging\FlashMessage;
use TYPO3\CMS\Core\Page\PageRenderer;
use TYPO3\CMS\Core\Utility\GeneralUtility;
use TYPO3\CMS\Core\Utility\VersionNumberUtility;
use TYPO3\CMS\Extbase\Mvc\Controller\ActionController;
use TYPO3\CMS\Extbase\Mvc\View\ViewInterface;
use TYPO3\CMS\Extbase\Object\ObjectManager;
use TYPO3\CMS\Extbase\Utility\LocalizationUtility;

/**
 * Controller for the mail templating service module
 */
class NewsletterController extends ActionController {
	/**
	 * DocHeaderComponent
	 *
	 * @var DocHeaderComponent
	 */
	protected $docHeaderComponent;

	/**
	 * @var TemplateRepository
	 */
	protected $templateRepository;

	/**
	 * Inject the TemplateRepository
	 *
	 * @param TemplateRepository $templateRepository
	 */
	public function injectTemplateRepository(TemplateRepository $templateRepository) {
		$this->templateRepository = $templateRepository;
	}

	/**
	 * @var FrontendUserGroupRepository
	 */
	protected $frontendUserGroupRepository;

	/**
	 * Inject the FrontendUserGroupRepository
	 *
	 * @param FrontendUserGroupRepository $templ$frontendUserGroupRepositoryateRepository
	 */
	public function injectFrontendUserGroupRepository(FrontendUserGroupRepository $frontendUserGroupRepository) {
		$this->frontendUserGroupRepository = $frontendUserGroupRepository;
	}

	/**
	 * @var LayoutRepository
	 */
	protected $layoutRepository;

	/**
	 * Inject the LayoutRepository
	 *
	 * @param LayoutRepository $layoutRepository
	 */
	public function injectLayoutRepository(LayoutRepository $layoutRepository) {
		$this->layoutRepository = $layoutRepository;
	}

	/**
	 * @var \SGalinski\SgMail\Session\PhpSession
	 */
	protected $session;

	/**
	 * @param ViewInterface $view
	 */
	public function initializeView(ViewInterface $view) {
		parent::initializeView($view);
		$pageRenderer = GeneralUtility::makeInstance(PageRenderer::class);
		$pageRenderer->addJsInlineCode(
			'typo3_version', 'TYPO3.version='
			. VersionNumberUtility::convertVersionNumberToInteger(VersionNumberUtility::getCurrentTypo3Version())
			. ';'
		);
		if (version_compare(VersionNumberUtility::getCurrentTypo3Version(), '9.0.0', '<')) {
			$pageRenderer->loadExtJS();
		}
		$view->assign('controller', 'Newsletter');
	}

	/**
	 * Show template Selection and enable content input + mail preview
	 *
	 * @param array $parameters
	 * @throws \InvalidArgumentException
	 * @throws \UnexpectedValueException
	 * @throws \TYPO3\CMS\Extbase\Mvc\Exception\StopActionException
	 * @throws \TYPO3\CMS\Extbase\Mvc\Exception\UnsupportedRequestTypeException
	 * @throws \BadFunctionCallException
	 * @throws \TYPO3\CMS\Core\Cache\Exception\NoSuchCacheException
	 * @throws \TYPO3\CMS\Extbase\Mvc\Exception\NoSuchArgumentException
	 */
	public function indexAction(array $parameters = []) {
		if ($this->request->hasArgument('message')) {
			$this->addFlashMessage($this->request->getArgument('message'), '', FlashMessage::INFO);
		}

		$pid = (int) GeneralUtility::_GP('id');

		if (!($this->session instanceof PhpSession)) {
			$this->session = $this->objectManager->get(PhpSession::class);
			$this->session->setSessionKey('sg_mail_controller_session');
		} else {
			$this->session->setSessionKey('sg_mail_controller_session');
		}

		if ($this->request->hasArgument('controller')) {
			$this->session->setDataByKey('mode', $this->request->getArgument('controller'));
		}
		if ($this->session->getDataByKey('mode') !== BackendService::BACKEND_MODE_QUEUE_NEWSLETTER) {
			$this->redirect('index', $this->session->getDataByKey('mode'));
		}

		$removeExtensionKeys = [];
		$registerArray = BackendService::getNonBlacklistedTemplates($pid);
		foreach ($registerArray as $extensionKey => $extensionTemplates) {
			if (\count($extensionTemplates) === 0) {
				$removeExtensionKeys[] = $extensionKey;
			}
		}

		foreach ($removeExtensionKeys as $extensionKey) {
			unset($registerArray[$extensionKey]);
		}
		reset($registerArray);

		// if no extensions are registered, redirect to empty action
		if (!\is_array($registerArray) || \count($registerArray) <= 0) {
			$this->forward('empty');
		}
		$arguments = $this->request->getArguments();

		$languages = BackendService::getLanguages();

		// if no template & extensionKey is selected look for them in the session
		if ($parameters['selectedTemplate'] === NULL || $parameters['selectedTemplate'] === ''
			|| $parameters['selectedLanguage'] === NULL) {
			if (
				$this->session->getDataByKey('selectedTemplate') !== NULL &&
				$this->session->getDataByKey('selectedExtension') !== NULL &&
				$this->session->getDataByKey('selectedLanguage') !== NULL &&
				!MailTemplateService::isTemplateBlacklisted(
					$this->session->getDataByKey('selectedExtension'),
					$this->session->getDataByKey('selectedTemplate'),
					$pid
				)
			) {
				$parameters['selectedTemplate'] = $this->session->getDataByKey('selectedTemplate');
				$parameters['selectedExtension'] = $this->session->getDataByKey('selectedExtension');
				$parameters['selectedLanguage'] = $this->session->getDataByKey('selectedLanguage');
			} else {
				// if nothing is set in the session, use the first entries from the register array
				$parameters['selectedExtension'] = key($registerArray);
				$parameters['selectedTemplate'] = key($registerArray[$parameters['selectedExtension']]);
				$parameters['selectedLanguage'] = $languages[0]['isocode'];
			}
		}

		// store selected template & extension key in the session
		$this->session->setDataByKey('selectedTemplate', $parameters['selectedTemplate']);
		$this->session->setDataByKey('selectedExtension', $parameters['selectedExtension']);
		$this->session->setDataByKey('selectedLanguage', $parameters['selectedLanguage']);

		$templatesFromDb = BackendService::getSelectedTemplates(
			$parameters['selectedExtension'], $parameters['selectedTemplate'], [BackendService::getLanguageByISOCode($parameters['selectedLanguage'])],
			$pid
		);

		// if no templates are in the db, get the default from the files
		$templates = [];
		$subject = $arguments['subject'] ?? $registerArray[$parameters['selectedExtension']][$parameters['selectedTemplate']]['subject'];
		/** @var  Template $template */
		foreach ($templatesFromDb as $key => $template) {
			if ($template === NULL) {
				$defaultTemplatePath = $registerArray[$parameters['selectedExtension']][$parameters['selectedTemplate']]['templatePath'];
				$fallbackTemplateFile = $defaultTemplatePath . 'template.html';

				$templateFromFile = new Template();
				$templateFromFile->setLanguage($key);
				$templates[$key] = $templateFromFile;
				if (\strpos($subject, 'LLL:') === 0) {
					/** @noinspection PhpMethodParametersCountMismatchInspection */
					$subject = LocalizationUtility::translate($subject, NULL, NULL, $key);
				}
				$templateFromFile->setSubject($subject);
				$subject = $arguments['subject'] ?? $registerArray[$parameters['selectedExtension']][$parameters['selectedTemplate']]['subject'];
				if ($registerArray[$parameters['selectedExtension']][$parameters['selectedTemplate']]['templateContent']) {
					$templateFromFile->setContent(
						$registerArray[$parameters['selectedExtension']][$parameters['selectedTemplate']]['templateContent']
					);
				} elseif (file_exists($fallbackTemplateFile)) {
					$templateFromFile->setContent(file_get_contents($fallbackTemplateFile));
				}
			} else {
				$templates[$key] = $template;
				$template->setIsOverwritten(TRUE);
			}

			if (!empty($arguments['content'])) {
				$templates[$key]->setContent($arguments['content']);
			}
			if (!empty($arguments['cc'])) {
				$templates[$key]->setCc($arguments['cc']);
			}
			if (!empty($arguments['bcc'])) {
				$templates[$key]->setBcc($arguments['bcc']);
			}
			if (!empty($arguments['fromName'])) {
				$templates[$key]->setFromName($arguments['fromName']);
			}
			if (!empty($arguments['fromMail'])) {
				$templates[$key]->setFromMail($arguments['fromMail']);
			}
			if (!empty($arguments['replyTo'])) {
				$templates[$key]->setReplyTo($arguments['replyTo']);
			}
			if (!empty($arguments['layout'])) {
				$templates[$key]->setLayout($arguments['layout']);
			}
			if (!empty($arguments['subject'])) {
				$templates[$key]->setSubject($arguments['subject']);
			}
		}

		// calculating optimal column width for the view
		$colspace = 4;
		$templateCount = \count($templates);
		if ($templateCount % 2 === 0 && $templateCount <= 4) {
			$colspace = 6;
		} elseif ($templateCount === 1) {
			$colspace = 12;
		}

		$this->view->assign('colspace', $colspace);
		$this->view->assign('languageTemplates', $templates);
		$this->view->assign('languageLabels', BackendService::getLanguageLabels($languages));
		$this->view->assign('templates', $registerArray);

		$registerService = $this->objectManager->get(RegisterService::class);
		$this->view->assign(
			'isManual', $registerService->isManuallyRegisteredTemplate($parameters['selectedTemplate'])
		);

		$templateDescription = $registerArray[$parameters['selectedExtension']][$parameters['selectedTemplate']]['description'];

		if (\is_array($templateDescription)) {
			if ($languages[0]['isocode']) {
				$templateDescription = $templateDescription[$languages[0]['isocode']];
			}
		} elseif (\strpos($templateDescription, 'LLL:') !== FALSE) {
			$templateDescription = LocalizationUtility::translate(
				$templateDescription, $parameters['selectedExtension']
			);
		}

		// create doc header component
		$pageUid = (int) GeneralUtility::_GP('id');
		$pageInfo = BackendUtility::readPageAccess($pageUid, $GLOBALS['BE_USER']->getPagePermsClause(1));

		if ($pageInfo && (int) $pageInfo['is_siteroot'] === 1) {

			$this->view->assign('templateDescription', $templateDescription);
			$this->view->assign(
				'selectedTemplate', $registerArray[$parameters['selectedExtension']][$parameters['selectedTemplate']]
			);
			$this->view->assign('selectedTemplateKey', $parameters['selectedTemplate']);
			$this->view->assign('selectedExtensionKey', $parameters['selectedExtension']);
			$this->view->assign('mode', 'editor');
			$layouts = $this->layoutRepository->findByPidForModule($pageUid);
			$layoutOptions = [
				0 => LocalizationUtility::translate('backend.layout.default', 'SgMail'),
				-1 => LocalizationUtility::translate('backend.layout.none', 'SgMail')
			];
			foreach ($layouts as $layout) {
				$layoutOptions[(int) $layout['uid']] = $layout['name'];
			}
			$this->view->assign('layoutOptions', $layoutOptions);
		} else {
			$this->view->assign('pages', BackendService::getPages());
		}

		$this->docHeaderComponent = GeneralUtility::makeInstance(DocHeaderComponent::class);
		if ($pageInfo === FALSE) {
			$pageInfo = ['uid' => $pageUid];
		}
		$this->docHeaderComponent->setMetaInformation($pageInfo);
		BackendService::makeButtons($this->docHeaderComponent, $this->request);

		$this->view->assign('docHeader', $this->docHeaderComponent->docHeaderContent());
		$this->view->assign('typo3Version', VersionNumberUtility::convertVersionNumberToInteger(TYPO3_version));
		/** @noinspection PhpInternalEntityUsedInspection */
		$beUserMail = (!empty($arguments['emailAddress'])?$arguments['emailAddress']:$GLOBALS['BE_USER']->user['email']);
		$this->view->assign('beUserMail', $beUserMail);

		// get the default language label and pass it to the view
		$pageTsConfig = BackendUtility::getPagesTSconfig($pageUid);
		if (VersionNumberUtility::convertVersionNumberToInteger(TYPO3_version) >= 8000000) {
			$defaultLanguageLabel = LocalizationUtility::translate(
				'LLL:EXT:lang/Resources/Private/Language/locallang_mod_web_list.xlf:defaultLanguage'
			);
		} else {
			$defaultLanguageLabel = LocalizationUtility::translate(
				'lll:EXT:lang/locallang_mod_web_list.xlf:defaultLanguage'
			);
		}

		if (isset($pageTsConfig['mod.']['SHARED.']['defaultLanguageLabel'])) {
			$defaultLanguageLabel = $pageTsConfig['mod.']['SHARED.']['defaultLanguageLabel'];
		}

		$this->view->assign('defaultLanguageLabel', $defaultLanguageLabel);
		$this->view->assign('selectedLanguage', $parameters['selectedLanguage']);

		$arguments = $this->request->getArguments();
		if (!isset($arguments['selectedGroups']) || !is_array($arguments['selectedGroups'])) {
			$selectedGroups = [];
		} else {
			$selectedGroups = $arguments['selectedGroups'];
		}
		$selectedGroupsTree = json_encode($this->frontendUserGroupRepository->getFrontendUserGroupsTree($selectedGroups), TRUE);

		$this->view->assign('selectedGroupsTree', $selectedGroupsTree);
	}

	/**
	 * send a test email to a given address
	 * redirect to index action
	 *
	 * @param array $parameters
	 * @throws \InvalidArgumentException
	 * @throws \TYPO3\CMS\Extbase\Persistence\Exception\IllegalObjectTypeException
	 * @throws \TYPO3\CMS\Extbase\Mvc\Exception\StopActionException
	 * @throws \TYPO3\CMS\Extbase\Mvc\Exception\UnsupportedRequestTypeException
	 * @throws \TYPO3\CMS\Extbase\Persistence\Exception\UnknownObjectException
	 * @throws \BadFunctionCallException
	 * @throws \TYPO3\CMS\Core\Cache\Exception\NoSuchCacheException
	 */
	public function sendTestMailAction(array $parameters = []) {
		$arguments = [];
		if ($parameters['selectedExtension']) {
			$arguments['selectedExtension'] = $parameters['selectedExtension'];
		}
		if ($parameters['selectedTemplate']) {
			$arguments['selectedTemplate'] = $parameters['selectedTemplate'];
		}

		foreach ((array) $parameters['templates'] as $parameter) {
			$ccAddresses = GeneralUtility::trimExplode(',', $parameter['cc']);
			if (\count($ccAddresses) > 0) {

				foreach ($ccAddresses as $ccAddress) {
					if (!filter_var($ccAddress, FILTER_VALIDATE_EMAIL) && trim($ccAddress) !== '') {
						$message = LocalizationUtility::translate('backend.error_cc', 'sg_mail');
						$this->addFlashMessage($message, '', FlashMessage::WARNING);

						$this->redirect('index', NULL, NULL, $arguments);
					}
				}
			}

			$bccAddresses = GeneralUtility::trimExplode(',', $parameter['bcc']);
			if (\count($bccAddresses) > 0) {
				foreach ($bccAddresses as $bccAddress) {
					if (!filter_var($bccAddress, FILTER_VALIDATE_EMAIL) && trim($bccAddress) !== '') {
						$message = LocalizationUtility::translate('backend.error_bcc', 'sg_mail');
						$this->addFlashMessage($message, '', FlashMessage::WARNING);

						$this->redirect('index', NULL, NULL, $arguments);
					}
				}
			}
		}

		$objectManager = GeneralUtility::makeInstance(ObjectManager::class);
		/** @var \SGalinski\SgMail\Service\MailTemplateService $mailTemplateService */
		$mailTemplateService = $objectManager->get(
			MailTemplateService::class, $parameters['selectedTemplate'], $parameters['selectedExtension']
		);
		$mailIsSend = FALSE;

		$mailTemplateService->setSubject($parameter['subject']);
		$mailTemplateService->setReplyToAddress($parameter['replyTo']);
		$mailTemplateService->setTemplateName($parameters['selectedTemplate']);
		$mailTemplateService->setExtensionKey($parameters['selectedExtension']);
		$mailTemplateService->setFromAddress($parameter['fromMail']);


		if (!$this->request->getArgument('sendRealEmails')) {
			// Send test emails
			$message = LocalizationUtility::translate('backend.success_mail', 'sg_mail');
			foreach ((array) $parameters['templates'] as $key => $template) {
				$mailTemplateService->setLanguage($key);
				$mailTemplateService->setToAddresses($parameters['emailAddress']);
				$mailTemplateService->setOverwrittenEmailBody($parameter['content']);
				$mailTemplateService->setMarkers([
					'user' => [
						'username' => 'username@example.com',
						'name' => 'Max Mustermann',
						'title' => 'Dr.',
						'first_name' => 'Max',
						'last_name' => 'Mustermann',
						'address' => 'Berlin, Germany',
						'zip' => '89073',
						'city' => 'Ulm',
						'country' => 'Germany',
						'telephone' => '491535555555',
						'email' => 'max.mustermann@example.com',
						'company' => 'Muster GmbH',
						'www' => '	www.example.com',
					]
				]);
				$mailIsSend = $mailTemplateService->sendEmail(TRUE);
			}

		} else {
			$errorRecipients = [];
			$message = LocalizationUtility::translate('backend.success_mail_queue', 'sg_mail');
			// Send real emails

			// get recipients
			if (!is_array($parameters['selectedGroups'])) {
				$selectedGroups = [];
			} else {
				$selectedGroups = $parameters['selectedGroups'];
			}
			$recipients = BackendService::getRecipientsByGroups($selectedGroups);


			foreach ((array) $parameters['templates'] as $key => $template) {
				$mailTemplateService->setLanguage($key);
				$mailTemplateService->setOverwrittenEmailBody($parameter['content']);
				$mailTemplateService->setBccAddresses($parameter['bcc']);
				$mailTemplateService->setCcAddresses($parameter['cc']);


				foreach ($recipients as $recipient) {
					try {
						$mailTemplateService->setToAddresses($recipient['email']);
						$mailTemplateService->setMarkers(['user' => $recipient]);
						// no real error handling here, one must check the MailQueue
						$mailTemplateService->sendEmail(FALSE);
					} catch (\Exception $e) {
						// Invalid email address could not be loaded to queue
						$errorRecipients[] = $recipient['uid'] . ' - '
							. $recipient['email'] . ' (' . $e->getMessage() . ')';
					}
				}
				$mailIsSend = TRUE;
			}
			if (count($errorRecipients) > 0) {
				$errorMessage = LocalizationUtility::translate('backend.newsletter.failure_mail', 'sg_mail');
				foreach ($errorRecipients as $errorRecipient) {
					$errorMessage .= "$errorRecipient\n";
				}
				$this->addFlashMessage($errorMessage, '', FlashMessage::ERROR);
			}
		}
		if ($mailIsSend) {
			$this->addFlashMessage($message, '', FlashMessage::OK);
		} else {
			$message = LocalizationUtility::translate('backend.failure_mail', 'sg_mail');
			$this->addFlashMessage($message, '', FlashMessage::ERROR);
		}
		$arguments['selectedGroups'] = $parameters['selectedGroups'];
		$arguments = array_merge($arguments, $parameter);
		$arguments['emailAddress'] = $parameters['emailAddress'];
		$this->redirect('index', NULL, NULL, $arguments);
	}

	/**
	 * show a notice when no extension is registered
	 */
	public function emptyAction() {

	}
}
